//===============================================================
// HPCALC: written by Scott T. Schad, Copyright 2001
//
// An RPN calculator based on the venerable HP-41
// Written in PocketC for Palm OS, created by
// Jeremy Dewey and Kevin Cao (www.orbworks.com)
//
// Requires Mathlib.prc
// Without warranty, for non-commercial applications
//
// Please donate through www.paypal.com to: petrologic@yahoo.com
//===============================================================
// Overview: 
//
// I wrote HPCALC because I wanted a small, fast, RPN calculator
// for my Palm handheld, something like the old HP41.  Other
// calculators for the Palm go overboard with too many features
// and shifted functions.  I kept HPCALC as simple as possible
// while still providing core engineering functions.
//
// Maybe HPCALC will help you learn Jeremy Dewey's delightful
// PocketC programming language!
//
//===============================================================
// How it works:
//
// The stack is four levels deep (x=top,y,z,t=bottom)
//
// ENTER pushes numbers through the x>y>z>t registers
// ENTER leaves a phantom copy in x: if the following key
//   is a number it replaces x, if an operation key it uses x
// To ensure this behavior, we track the status of the ENTER key
//   so a keypress knows if the last one=ENTER
//
// If you explicity DROP x or combine x,y in a calculation,
//   then a t>z>y copy occurs.  Because T retains its value,
//   it acts as a storage register.
// Successive DROPs copy the T value into all registers.
// To clear all four registers, type a zero and press ENTER three times
//
// Forgive my engineering bias, but trig functions use degrees.
//   To convert to radians, multiply by pi/180
//
// To enter an exponent:
//   key in numbers, then EEX, then the exponent
// To enter a signed number and a signed exponent:
//   key in numbers, CHS, then EEX, then CHS, then the exponent
//
//===============================================================
//Examples: algebraic entry vs Reverse Polish Notation
//
// Algebraic: 2 + 3 = 
// RPN: 2 ENTER 3 +
//
// Algrebraic: 10 / 5 =
// RPN: 10 ENTER 5 /
//
// Algebraic: log(10)
// RPN: 10 log
//
// Algebraic: 2 y^x 3 =
// RPN: 2 ENTER 3 y^x
//
// Algebraic: sqrt( (.62/.15^2.15)*.04 / 9 )
// RPN: .62 ENTER .15 ENTER 2.15 y^x / .04 * 9 / sqrt
//
//===============================================================

@cid "STSc";
@ver "1.0";
@name "HPCALC";
@dbname "HPCALC";

// declare some variables
int xi,yi,xpen,ypen,xcol,yrow,opcode;
string strname,xreg,yreg,zreg,treg,tempx,lastkey,temp;
string opname,dispx;

// define routines to manipulate the stack
pushx() {
    // X Y Z T register sequence
    treg=zreg;
    zreg=yreg;
    yreg=xreg;
}

dropx() {
	xreg=yreg;
	yreg=zreg;
	zreg=treg;
    // let values persist in T register
}

dropy() {
	yreg=zreg;
	zreg=treg;
    // let values persist in T register
}

// parses pen taps
parse() {
	if (xpen<80)
		if (xpen<42)
			xcol=1;
		else
			xcol=2;
	else
		if (xpen<118)
			xcol=3;
		else
			xcol=4;
	if (ypen<94)
		if (ypen<49)
			if (ypen<34)
				yrow=1;
			else
				yrow=2;
		else
			if (ypen<64)
				yrow=3;
			else
				if (ypen<79)
					yrow=4;
				else
					yrow=5;
	else
		if (ypen<124)
			if (ypen<109)
				yrow=6;
			else
				yrow=7;
		else
			if (ypen<139)
				yrow=8;
			else
				yrow=9;
				
	// at this point you know coords of key,
	// so assign a numeric opcode to it
	switch(yrow) {
		case 1:
			switch(xcol) {
				case 1:
					strname = "1/x";
					opcode=1;
					break;
				case 2:
    				strname = "sqrt";
    				opcode=2;
    				break;
    			case 3:
    				strname = "log";
    				opcode=3;
    				break;
    			case 4:
    				strname = "ln";
    				opcode=4;
    				break;
    		}
    		break;
    	case 2:
    		switch(xcol) {
    			case 1: 
    				strname = "y^x";
    				opcode=5;
    				break;
	    		case 2:
    				strname = "x^2";
    				opcode=6;
    				break;
    			case 3:
    				strname = "10^x";
    				opcode=7;
    				break;
    			case 4:
    				strname = "e^x";
    				opcode=8;
    				break;
    		}
    		break;
    	case 3:
    		switch(xcol) {
    			case 1:
    				strname = "pi";
    				opcode=9;
    				break;
	    		case 2:
    				strname = "sin";
    				opcode=10;
    				break;
    			case 3:
    				strname = "cos";
    				opcode=11;
    				break;
    			case 4:
    				strname = "tan";
    				opcode=12;
    				break;
    		}
    		break;
    	case 4:
    		switch(xcol) {
    			case 1:
    				strname = "swap";
    				opcode=13;
    				break;
    			case 2:
    				strname = "asin";
    				opcode=14;
    				break;
       			case 3:
    				strname = "acos";
    				opcode=15;
    				break;
    			case 4:
    				strname = "atan";
    				opcode=16;
    				break;
    		}
    		break;
    	case 5:
    		switch(xcol) {
    			case 1:
    				strname = "E N T E R";
    				opcode=17;
    				break;
    			case 2:
    				strname = "E N T E R";
    				opcode=18;
    				break;
    			case 3:
    				strname = "chs";
    				opcode=19;
    				break;
    			case 4:
    				strname = "drop";
    				opcode=20;
    				break;
    		}
    		break;
    	case 6:
    		switch(xcol) {
    			case 1:
    				strname = "-";
    				opcode=21;
    				break;
    			case 2:
    				strname = "7";
    				opcode=22;
    				break;
    			case 3:
    				strname = "8";
    				opcode=23;
    				break;
    			case 4:
    				strname = "9";
    				opcode=24;
    				break;
    		}
    		break;
    	case 7:
    		switch(xcol) {
    			case 1:
    				strname = "+";
    				opcode=25;
    				break;
    			case 2:
    				strname = "4";
    				opcode=26;
    				break;
    			case 3:
    				strname = "5";
    				opcode=27;
    				break;
    			case 4:
    				strname = "6";
    				opcode=28;
    				break;
    		}
    		break;
    	case 8:
    		switch(xcol) {
    			case 1:
    				strname = "*";
    				opcode=29;
    				break;
    			case 2:
    				strname = "1";
    				opcode=30;
    				break;
    			case 3:
    				strname = "2";
    				opcode=31;
    				break;
    			case 4:
    				strname = "3";
    				opcode=32;
    				break;
    		}
    		break;
    	case 9:
    		switch(xcol) {
    			case 1:
    				strname = "/";
    				opcode=33;
    				break;
    			case 2:
    				strname = "0";
    				opcode=34;
    				break;
    			case 3:
    				strname = ".";
    				opcode=35;
    				break;
    			case 4:
    				strname = "eex";
    				opcode=36;
    				break;
    		}
    	}

	if (xpen<40) {
		if (ypen<13) {
			opcode=0;
			strname="noop";
		}
	}
    return strname;
}

// start of main program
main(){
	
	// Is mathlib present?
	if (mathlib()==0) {
		alert("You need MathLib installed!");
		exit();
	}

	graph_on();
    clearg();
 	title("HPCALC");
   
   	// initialize variables
	xreg="0";
	yreg="0";
	zreg="0";
	treg="0";   
	tempx="";
	lastkey="number";	// number, operation, or ENTER

	// draw the calculator
	xi=23;
	do {
		yi=24;
		do {
			if (yi!=84){
				frame(1,xi-16,yi-3,xi+16,yi+8,3);
			}
			else
				if (xi<70)
					frame(1,7,yi-3,77,yi+8,3);
				else
					frame(1,xi-16,yi-3,xi+16,yi+8,3);
			yi=yi+15;
		} while (yi<150);

		xi=xi+38;
	} while (xi<180);

	// label the keys
    textalign(11);
	text(23,25,"1/x");
    text(61,25,"sqrt");
    text(99,25,"log");
    text(137,25,"ln");
    
    text(23,40,"y^x");
    text(61,40,"x^2");
    text(99,40,"10^x");
    text(137,40,"e^x");
    
    text(23,55,"pi");
    text(61,55,"sin");
    text(99,55,"cos");
    text(137,55,"tan");
    
    text(23,70,"swap");
    text(61,70,"asin");
    text(99,70,"acos");
    text(137,70,"atan");
    
    text(42,85,"E N T E R");
    text(99,85,"chs");
    text(137,85,"drop");
    
    text(23,100,"-");
    text(61,100,"7");
    text(99,100,"8");
    text(137,100,"9");
    
    text(23,115,"+");
    text(61,115,"4");
    text(99,115,"5");
    text(137,115,"6");
    
    text(23,130,"*");
    text(61,130,"1");
    text(99,130,"2");
    text(137,130,"3");
    
    text(23,145,"/");
    text(61,145,"0");
    text(99,145,".");
    text(137,145,"eex");

	textalign(0);
	
	// display date & time at startup
 	text(50,2,date(2)+" "+time(1));

	frame(1,1,1,159,159,0);
	frame(1,2,2,158,158,0);
	
	// say hi
	alert("HPCALC (c) 2001\nby Scott T. Schad.\n\npetrologic@yahoo.com");

	//now trap pen events
	do {
		waitp();
		xpen=(penx());
		ypen=(peny());
		
		// wipe display area
		rect(0,50,2,157,13,0);
		// parse the key
		tempx=parse();

        if (strstr("9876543210.e", tempx, 0)!=-1) {
            // new key is a number
            // how you handle it depends on
            // the previous key pressed
            if (lastkey=="number") {
				if (strlen(xreg)<15) {
					if (xreg=="0")
						xreg=tempx;
					else
               			xreg=xreg + tempx;
 
                    lastkey="number";
                }
            }
            if (lastkey=="ENTER") {
            	xreg=tempx;
                lastkey="number";
            }
			if (lastkey=="operation") {
            	pushx();
                xreg=tempx;
                lastkey="number";
            }
        }
		else{
			// new key is an operation or ENTER
			// figure out what to do by the opcode
		switch(opcode) {
				case 1:	// "1/x"
					if (xreg=="0") {
						alert ("Division by zero");
						break;
					}
					xreg=(1/((float)xreg))+"";
	        		lastkey="operation";
	        		break;
				case 2: // "sqrt"
                    xreg=sqrt((float)xreg)+"";
	        		lastkey="operation";
                    break;
    			case 3:	// "log"
    				xreg=log10((float)xreg)+"";
	        		lastkey="operation";
    				break;
    			case 4: // "ln"
    				xreg=log((float)xreg)+"";
	        		lastkey="operation";
    				break;
    			case 5: // "y^x"
    				xreg=pow((float)yreg,(float)xreg)+"";
                    dropy();
	        		lastkey="operation";
    				break;
	    		case 6: // "x^2"
    				xreg=(((float)xreg)*((float)xreg))+"";
	        		lastkey="operation";
    				break;
    			case 7: // "10^x"
    				xreg=pow(10.0,(float)xreg)+"";
	        		lastkey="operation";
    				break;
    			case 8: // "e^x"
    				xreg=exp((float)xreg)+"";
	        		lastkey="operation";
    				break;
    			case 9: // "pi"
    				pushx();
					xreg=3.1415926535898+"";
	        		lastkey="operation";
    				break;
	    		case 10: // "sin"
    				xreg=sin((float)xreg*3.1415926535898/180)+"";
	        		lastkey="operation";
    				break;
    			case 11: // "cos"
    				xreg=cos((float)xreg*3.1415926535898/180)+"";
	        		lastkey="operation";
    				break;
    			case 12: // tan"
    				xreg=tan((float)xreg*3.1415926535898/180)+"";
	        		lastkey="operation";
    				break;
    			case 13: // "swap"
    				if (yreg!=""){
                       	temp=xreg;
                       	xreg=yreg;
                       	yreg=temp;
	        			lastkey="operation";
                       	break;
                    }
                    else
    					break;
    			case 14: // "asin"
    				xreg=asin((float)xreg)*180/3.1415926535898+"";
	        		lastkey="operation";
    				break;
       			case 15: // "acos"
    				xreg=acos((float)xreg)*180/3.1415926535898+"";
	        		lastkey="operation";
    				break;
    			case 16: // "atan"
    				xreg=atan((float)xreg)*180/3.1415926535898+"";
	        		lastkey="operation";
    				break;
    			case 17: // "E N T E R"
    			case 18: // "E N T E R"
                    if (xreg!="") {
						
						dispx=format((float)xreg,14);

						// trim trailing zeros
						while (strright(dispx,1)=="0") {
							dispx=strleft(dispx,strlen(dispx)-1);
						}
					
						xreg=dispx;
						
                       	pushx();
                       	lastkey="ENTER";
                    	break;
                    }
                    else
    					break;
    			case 19: // "chs"
    				if(strlen(xreg)==0)
    					break;
    				if (strright(xreg,1)=="-"){
    					xreg=strleft(xreg,strlen(xreg)-1);
    					break;
    				}
    				if(strleft(xreg,1)=="-")
    					if (strright(xreg,1)!="e"){
    						xreg=strright(xreg,strlen(xreg)-1);
    						break;
    					}
    					else {
    						xreg=xreg+"-";
    						break;
    					}
    				if (strright(xreg,1)=="e")
    					xreg=xreg+"-";
    				else
    					xreg="-"+xreg;	
    				break;
    			case 20: // "drop"
    				dropx();
	        		lastkey="operation";
    				break;
    			case 21: // "-"
    				xreg=((float)yreg)-((float)xreg)+"";
                    dropy();
	        		lastkey="operation";
    				break;
    			case 25: // "+"
    				xreg=((float)yreg+(float)xreg)+"";
                    dropy();
	        		lastkey="operation";
    				break;
    			case 29: // "*"
    				xreg=((float)yreg)*((float)xreg)+"";
                    dropy();
	        		lastkey="operation";
    				break;
    			case 33: // "/"
					if (xreg=="0") {
						alert ("Division by zero");
						break;
					}
                    xreg=((float)yreg)/((float)xreg)+"";
                    dropy();
	        		lastkey="operation";
    				break;
    			case 36: // "eex"
               		if (strstr(xreg,"e",1)<1) {
                   		// avoid multiple exponents	
                    	xreg=xreg+"e";
                    	lastkey="number";
  	                   	break;
               		}
    		} // switch
    	} //else

		if (lastkey=="operation")
			dispx=format((float)xreg,14);
		else
			dispx=xreg;

		// for frisky pen taps in the display area,
		// just redisplay the system date and time
		if (ypen<12) {
			rect(0,50,2,157,13,0);
 			dispx=date(2)+" "+time(1);
 		}
			
		// trim trailing zeros
		if (strstr(dispx,".",0)>0) {
			while (strright(dispx,1)=="0") {
				dispx=strleft(dispx,strlen(dispx)-1);
			}
		}
		text(50,2,dispx);
	} while (ypen<160);
	wait();
	graph_off();
	
}

// that's all folks!